{-# LANGUAGE ScopedTypeVariables #-}

module Kronecker where

import Data.List
import Data.Ratio
import Data.Array
import Data.Maybe
import RandomState
import qualified Control.Monad.State as MS
import System.Random
import Automata

data InitGraph = C2 | C3 | C4 | R3 | R4 | Y4 | S4 | Y4R3 | C5 | Y5 | Y5R3 | S5 | R5 deriving (Show,Read,Eq,Ord)
--Make Y4=S4 part of the Read instance, Put in S3=C3, put in Y4R3
type KrMatrix a = Array (Int, Int) a

kroneckerByInit :: Bool -> Int -> InitGraph -> Int -> Rational -> Rational -> Rational -> RandomState Automaton
kroneckerByInit initTrans sigma gName k a b f =
                kroneckerAutomaton initTrans sigma (initToGraph gName) k a b f

initToGraph :: InitGraph -> KrMatrix Int
initToGraph gName = 
		case gName of
                          C2 -> c2
                          C3 -> c3
                          C4 -> c4
                          R3 -> r3
                          R4 -> r4
                          Y4 -> y4
                          S4 -> y4
                          Y5 -> y5
                          S5 -> s5
                          Y4R3 -> y4r3
                          Y5R3 -> y5r3
                          C5 -> c5
                          R5 -> r5

krCalcN :: InitGraph -> Int -> Int
krCalcN gName k = origSize ^ k
            where origSize = (fst $ snd bnds) - (fst $ fst bnds) + 1
                  bnds = bounds (initToGraph gName)
            
--Chain graphs.
c2 = array ((0,0),(1,1)) [((0,0),1),((0,1),1),
                          ((1,0),1),((1,1),1)]
c3 = array ((0,0),(2,2)) [((0,0),1),((0,1),1),((0,2),0),
                          ((1,0),1),((1,1),1),((1,2),1),
                          ((2,0),0),((2,1),1),((2,2),1)]
c4 = array ((0,0),(3,3)) [((0,0),1),((0,1),1),((0,2),0),((0,3),0),
                          ((1,0),1),((1,1),1),((1,2),1),((1,3),0),
                          ((2,0),0),((2,1),1),((2,2),1),((2,3),1),
                          ((3,0),0),((3,1),0),((3,2),1),((3,3),1)]
c5 = array ((0,0),(4,4)) [((0,0),1),((0,1),1),((0,2),0),((0,3),0),((0,4),0),
                          ((1,0),1),((1,1),1),((1,2),1),((1,3),0),((1,4),0),
                          ((2,0),0),((2,1),1),((2,2),1),((2,3),1),((2,4),0),
                          ((3,0),0),((3,1),0),((3,2),1),((3,3),1),((3,4),1),
                          ((4,0),0),((4,1),0),((4,2),0),((4,3),1),((4,4),1)]
--Ring graphs.
r3 = array ((0,0),(2,2)) [((0,0),1),((0,1),1),((0,2),1),
                          ((1,0),1),((1,1),1),((1,2),1),
                          ((2,0),1),((2,1),1),((2,2),1)]
r4 = array ((0,0),(3,3)) [((0,0),1),((0,1),1),((0,2),1),((0,3),0),
                          ((1,0),1),((1,1),1),((1,2),0),((1,3),1),
                          ((2,0),1),((2,1),0),((2,2),1),((2,3),1),
                          ((3,0),0),((3,1),1),((3,2),1),((3,3),1)]
r5 = array ((0,0),(4,4)) [((0,0),1),((0,1),1),((0,2),0),((0,3),0),((0,4),1),
                          ((1,0),1),((1,1),1),((1,2),1),((1,3),0),((1,4),0),
                          ((2,0),0),((2,1),1),((2,2),1),((2,3),1),((2,4),0),
                          ((3,0),0),((3,1),0),((3,2),1),((3,3),1),((3,4),1),
                          ((4,0),1),((4,1),0),((4,2),0),((4,3),1),((4,4),1)]
--Y graphs.
y4 = array ((0,0),(3,3)) [((0,0),1),((0,1),1),((0,2),0),((0,3),0),
                          ((1,0),1),((1,1),1),((1,2),1),((1,3),1),
                          ((2,0),0),((2,1),1),((2,2),1),((2,3),0),
                          ((3,0),0),((3,1),1),((3,2),0),((3,3),1)]
y5 = array ((0,0),(4,4)) [((0,0),1),((0,1),1),((0,2),0),((0,3),0),((0,4),0),
                          ((1,0),1),((1,1),1),((1,2),1),((1,3),0),((1,4),0),
                          ((2,0),0),((2,1),1),((2,2),1),((2,3),1),((2,4),1),
                          ((3,0),0),((3,1),0),((3,2),1),((3,3),1),((3,4),0),
                          ((4,0),0),((4,1),0),((4,2),1),((4,3),0),((4,4),1)]
--Star graphs. S4 is Y4.
s5 = array ((0,0),(4,4)) [((0,0),1),((0,1),1),((0,2),1),((0,3),1),((0,4),1),
                          ((1,0),1),((1,1),1),((1,2),0),((1,3),0),((1,4),0),
                          ((2,0),1),((2,1),0),((2,2),1),((2,3),0),((2,4),0),
                          ((3,0),1),((3,1),0),((3,2),0),((3,3),1),((3,4),0),
                          ((4,0),1),((4,1),0),((4,2),0),((4,3),0),((4,4),1)]
--Y-graphs with terminating ring. 
y4r3 = array ((0,0),(3,3)) [((0,0),1),((0,1),1),((0,2),0),((0,3),0),
                            ((1,0),1),((1,1),1),((1,2),1),((1,3),1),
                            ((2,0),0),((2,1),1),((2,2),1),((2,3),1),
                            ((3,0),0),((3,1),1),((3,2),1),((3,3),1)]
y5r3 = array ((0,0),(4,4)) [((0,0),1),((0,1),1),((0,2),0),((0,3),0),((0,4),0),
                            ((1,0),1),((1,1),1),((1,2),1),((1,3),0),((1,4),0),
                            ((2,0),0),((2,1),1),((2,2),1),((2,3),1),((2,4),1),
                            ((3,0),0),((3,1),0),((3,2),1),((3,3),1),((3,4),1),
                            ((4,0),0),((4,1),0),((4,2),1),((4,3),1),((4,4),1)]

kroneckerAutomaton :: Bool -> Int -> KrMatrix Int -> Int -> Rational -> Rational -> Rational -> RandomState Automaton
--initTrans forces an outbound transition on each character from the start state.
--Taking sigma = number of characters, k = Kronecker power, a = stochastic replacement for 1, b = stochastic replacement for 0, 
--f = acceptance density, and a generating adjacency matrix,  generates a Kronecker random automaton.
kroneckerAutomaton initTrans sigma arr k a b f = 
        do
          let abArr = array (bounds arr) [if x == 1 then (ind,a) else (ind,b) | ind <- (range $ bounds arr), let x = arr ! ind]
          let finArr = kroneckerPower abArr k
          let ((xMin,yMin),(xMax,yMax)) = bounds finArr
          let size = (xMax-xMin+1)
          graphs <- sequence $ replicate sigma (kroneckerGraph initTrans size finArr)
          finStates <- getRandomSubset [0..size-1] (floor (f * (fromIntegral size)))
          return $ automataCompose [0] graphs finStates

kroneckerGraph :: Bool -> Int -> KrMatrix Rational -> RandomState Graph
--initTrans forces an outbound transition on each character from the start state.
--size is the size of the graph, passed in so as not to be calculated multiple times
--arr is a stochastic adjancency matrix, with each index's value representing the probability of there being a directed edge there
--kroneckerGraph returns the graph after edge presence is randomly determined
kroneckerGraph initTrans size arr =
        do
          edges <- sequence [isPresent edge p | edge <- (range $ bounds arr), let p = arr ! edge]
          let edges' = fromJust $ sequence [x | x <- edges, isJust x]
          if initTrans
          then do
                 initDest <- getRandomR(0,size-1)
                 let initEdge = (0,initDest)
                 return (size,nub $ initEdge : edges')
          else return (size,edges')
        where 
          isPresent :: (Int,Int) -> Rational -> RandomState (Maybe (Int,Int))
          --With probability p, randomly determines if edge is present.
          isPresent edge p  = 
               do
                 r::Double <- getRandomR(0.0,1.0)
                 if r < (fromRational p)
                 then return $ Just edge
                 else return Nothing
          
--Returns the Kronecker product of arr and arr2.
kroneckerProduct :: Num a => KrMatrix a -> KrMatrix a -> KrMatrix a
kroneckerProduct arr arr2  = array ((0,0),(maxXProd, maxYProd)) assocList
        where
          (_,(maxX1,maxY1)) = bounds arr
          (_,(maxX2,maxY2)) = bounds arr2
          maxXProd = (maxX1+1) * (maxX2+1) -1
          maxYProd = (maxY1+1) * (maxY2+1) -1
          assocList = [((x,y), (arr!(i,k)) * (arr2!(j,l))) | 
                      i <- [0..maxX1], j <- [0..maxX2], k <- [0..maxY1], l <- [0..maxY2],
                      let x = (i*(maxX2+1))+j, let y = (k*(maxY2+1)) + l] --x and y are the new coordinates: insert better comment.

--Returns arr to the nth Kronecker power.
kroneckerPower :: Num a => KrMatrix a -> Int -> KrMatrix a
kroneckerPower arr 1 = arr
kroneckerPower arr n = kroneckerProduct arr (kroneckerPower arr (n-1))
