module Logit where

import Automata
import RandomState
import Data.List
import Data.Ratio
import qualified Control.Monad.State as MS
import Control.Monad.State hiding (State)
import Control.Monad 

type LogitState = (Int,Int)

cleanedLogitAutomaton :: Bool -> Int -> Int -> Rational -> Rational -> RandomState Automaton
cleanedLogitAutomaton initTrans sigma n r f = liftM clean_aut (logitAutomaton initTrans sigma n r f)

logitAutomaton :: Bool -> Int -> Int -> Rational -> Rational -> RandomState (PAutomaton LogitState)
--initTrans forces an outbound transition on each character from the start state.
--Taking sigma = number of characters, n = number of labels, r = transition density, f = acceptance density, generates a logit random automaton.
logitAutomaton initTrans sigma n r f = do
          let states = [(x,y) | x <- [0..n-1], y <- [x..n-1]]
          edges <- replicateM sigma (logitGraph initTrans states r)
          finStates <- getRandomSubset states (floor (f * (fromIntegral $ length states)))
          return $ Automaton sigma states [(0,0)] (graphsToRho edges) finStates
          

logitGraph :: Bool -> [LogitState] -> Rational -> RandomState (PGraph LogitState) 
--Generates a logit graph with length states states and length states*r edges.
logitGraph initTrans vertices r = 
      do let n = length vertices
         let numEdges = (floor ((fromIntegral n) * r))
         when (numEdges > 4*n) (fail "r too high")  --because there needs to be an else
         edges <- if initTrans then (initEdge numEdges) else (genEdges vertices numEdges [])
         return (vertices , edges)
      where
         initEdge :: Int -> RandomState [(LogitState,LogitState)]
         initEdge numEdges = do
                       v <- getRandomElem $ filter (\(x,y) -> x == 0 || y == 0) vertices
                       genEdges vertices (numEdges-1) [((0,0),v)]

genEdges :: [LogitState] -> Int -> [(LogitState,LogitState)] -> RandomState [(LogitState,LogitState)]
genEdges _ 0 currEdges = return currEdges
genEdges states count currEdges 
    | count <= 0 = return currEdges
    | otherwise = do
               u@(u1,u2) <- getRandomElem states
               v <- getRandomElem $ union (hasLabel u1) (hasLabel u2)
               let newEdge = (u,v)
               if newEdge `elem` currEdges
               then genEdges states count currEdges
               else genEdges states (count-1) (newEdge : currEdges)
    where
         hasLabel l = filter (\(x,y) -> l == x || l == y) states 
