module TabakovVardi where

import Automata
import RandomState
import Data.List
import Data.Ratio
import qualified Control.Monad.State as MS

tabakovVardiAutomaton :: Bool -> Int -> Int -> Rational -> Rational -> RandomState Automaton
--initTrans forces an outbound transition on each character from the start state.
--Taking sigma = number of characters, n = number of states, r = transition density, f = acceptance density, generates a TV random automaton.
tabakovVardiAutomaton initTrans sigma n r f = 
        do
          graphs <- sequence $ replicate sigma (tvGraph initTrans n r)
          finStates <- getRandomSubset [0..n-1] (floor (f * (fromIntegral n)))
          return $ automataCompose [0] graphs finStates


tvGraph :: Bool -> Int -> Rational -> RandomState Graph
--Generates a TV graph with n nodes and n*r edges.
tvGraph initTrans n r =
        do
          let numEdges = (floor ((fromIntegral n) * r))
          if numEdges > n*n then fail "r too high" else return () --because there needs to be an else
          edges <- if initTrans then (initEdge numEdges) else (genEdges n numEdges [])
          return (n,edges)
        where
          initEdge :: Int -> RandomState [(Int,Int)]
          initEdge numEdges =
                     do
                       v <- getRandomR (0,n-1)
                       genEdges n (numEdges-1) [(0,v)]

genEdges :: Int -> Int -> [(Int,Int)] -> RandomState [(Int,Int)]
genEdges _ 0 currEdges = return currEdges
genEdges n count currEdges 
         | count <= 0 = return currEdges
         | otherwise =
             do
               u <- getRandomR (0,n-1)
               v <- getRandomR (0,n-1)
               let newEdge = (u,v)
               if newEdge `elem` currEdges
               then genEdges n count currEdges
               else genEdges n (count-1) (newEdge : currEdges)
