module VertexCopying where

import Automata
import RandomState
import Data.List
import Data.Ratio
import Control.Monad.State hiding (State)
import qualified Control.Monad.State as MS

vertexCopyingAutomaton :: Bool -> Int -> Int -> Rational -> Rational -> Rational -> RandomState Automaton
--initTrans forces an outbound transition on each character from the start state.
--Taking sigma = number of characters, n = number of states, b = chance of copying r = transition density, f = acceptance density, generates a web-as-graph random automaton.
vertexCopyingAutomaton initTrans sigma n b r f = do
          graphs <- replicateM sigma (vcGraph initTrans n b r)
          finStates <- getRandomSubset [0..n-1] (floor (f * (fromIntegral n)))
          return $ automataCompose [0] graphs finStates

vcGraph :: Bool -> Int -> Rational -> Rational -> RandomState Graph
--Generates a VC graph with n nodes, b chance of copying, and n*r edges.
vcGraph initTrans n b r = do
          let numEdges = (floor ((fromIntegral n) * r))
          when (numEdges > n*n) $ fail "r too high"
          edges <- if initTrans then (initEdge b numEdges) else (genEdges n b numEdges [])
          return (n,edges)
        where
          initEdge ::  Rational -> Int -> RandomState [(Int,Int)]
          initEdge b numEdges = do
                       v <- getRandomR (0,n-1)
                       genEdges n b (numEdges-1) [(0,v)]

genEdges :: Int -> Rational -> Int -> [(Int,Int)] -> RandomState [(Int,Int)]
genEdges _ _ 0 currEdges = return currEdges
genEdges n b count currEdges 
         | count <= 0 = return currEdges
         | otherwise = do
               u <- getRandomR (0,n-1)
               v <- destination 
               let newEdge = (u,v)
               if newEdge `elem` currEdges 
               then genEdges n b count currEdges
               else genEdges n b (count-1) (newEdge : currEdges)
             where
               destination :: RandomState Int
               destination  = do
                        bb <- getRandomR (0.0,1.0) :: RandomState Double  
                        if (fromRational b) < bb
                        then getRandomR (0,n-1)
                        else do
                               let sources = nub $ map fst currEdges
                               src  <- getRandomElem sources 
                               getRandomElem [v | (u,v) <- currEdges, u == src] 
